structure base_lexer=
   struct
    type int = Int.int
    structure UserDeclarations =
      struct
(* this is an -*- sml -*- file, or near enough *)
 open base_tokens locn

  type extrastate = {nf : int, r : int ref, i: int ref,
                     stash : string ref,
                     rcopt : (int * int) option ref,
                     stracc : (string * string list * locn_point *
                               (string*string)) ref,
                     incommentp : int ref}
  (* mutable state argument to each rule is st=(nf,r,i,rcopt), where:
       - nf  is the number of the fragment being parsed
       - r   is the current row number
       - i   is the index of the first char on the current row
       - rcopt is the absolute line and character of the start of this fragment, if known
       - stracc has information for handling string literals
           (pfx,contents,loc,(ldelim,rdelim))
         * pfx is material that preceded the literal but abutted it (e.g.,
           input was f"foo"
         * contents is stuff that appeared between the delimiters, built up
           by consing (and is thus in reverse order)
         * loc is location of literal
         * (ldelim,rdelim) is observed left-delimiter and expected right-
           delimiter
  *)

  fun mkLoc (st as {rcopt, ...} : extrastate) s e
    = case !rcopt of
          NONE => Loc(s,e)
        | SOME(row,col) => Loc(rel_to_abs row col s, rel_to_abs row col e)

  fun getLoc (st as {nf,r,i,rcopt,...} : extrastate) pos lexeme
    = let val s = LocP(nf,!r,pos - !i)
          val e = LocP(nf,!r,pos + size lexeme - !i - 1)
      in
          mkLoc st s e
      end

  fun newstate nf =
      {nf = nf, r = ref 0, i = ref 0, rcopt = ref NONE, stash = ref "",
       stracc = ref ("", [], LocP(0,0,0) (* meaningless locn *), ("","")),
       incommentp = ref 0}

  (* processes location pragmas of the form (*#loc row col*), using
     them to determine the absolute position of fragments in the input
     stream. *)
  fun dolocpragma (st as {r,i,rcopt,...} : extrastate) pos yytext
    = let val s = Substring.full yytext
          val sr = Substring.dropl (not o Char.isDigit) s
          val sc = Substring.dropl (Char.isDigit) sr
      in
        r := 0;
        i := pos + size yytext;
        rcopt := SOME (valOf (Int.fromString(Substring.string sr)) - 1,
                       valOf (Int.fromString(Substring.string sc)) - 1)
      end
   type lexresult = unit base_token * locn
   fun eof (st:extrastate) = let
     val {incommentp,stash,stracc,rcopt,...} = st
     val comment = !incommentp <> 0
     val ok = not comment andalso List.null (#2 (!stracc)) andalso !stash = ""
     val loc = locn.locp (case !rcopt of
                              SOME a => LocA a |
                              _ => LocPEnd (#nf st))
   in
     if ok then (BT_EOI, loc)
     else if !stash <> "" then
       if String.sub(!stash,0) = #"\"" then
         (BT_StrLit{ldelim = "\"", contents = ""}, loc) before stash := ""
       else
         (BT_Ident (!stash), loc) before stash := ""
     else raise LEX_ERR
            ("eof/antiquote inside " ^
             (if comment then "comment" else "quote-delimited string"), loc)
   end

   fun check_char (tl as (tok, loc)) =
       case tok of
         BT_StrLit{contents = s,ldelim = "#\""} =>
           if size s = 1 then (tok, loc)
           else raise LEX_ERR ("bad character constant", loc)
        | _ => raise LEX_ERR ("Very bad character constant", loc)

fun cut_qid s =
    let
      open Substring
      val ss = full s
      fun equal x y = x = y
      val (uptodol, id) = splitr (not o equal #"$") ss
    in
      if size uptodol > 0 then
        let
          val beforedol = trimr 1 uptodol
          fun thyname c = c = #"_" orelse not (Char.isPunct c)
          val (first, thy) = splitr thyname beforedol
        in
          (string first, string thy ^ "$" ^ string id)
        end
      else ("", s)
    end

fun matching_strdelim s =
    case s of
        "\"" => "\""
      | "\194\171" => "\194\187" (* double guillemets *)
      | "\226\128\185" => "\226\128\186" (* single guillemets *)
      | "#\"" => "\""
      | _ => raise Fail "Bad string delimiter situation"

end (* end of user routines *)
exception LexError (* raised if illegal leaf action tried *)
structure Internal =
    struct

datatype yyfinstate = N of int
type statedata = {fin : yyfinstate list, trans: int Vector.vector}
(* transition & final state table *)
val tab = let
fun decode s k =
  let val k' = k + k
      val hi = Char.ord(String.sub(s, k'))
      val lo = Char.ord(String.sub(s, k' + 1))
  in hi * 256 + lo end
val s = [
 (0,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (1,256,
"\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\244\000\246\000\009\000\009\000\245\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\244\000\076\000\243\000\241\000\167\000\076\000\076\000\077\
\\000\140\000\076\000\076\000\076\000\076\000\076\000\076\000\076\
\\000\133\000\128\000\128\000\128\000\128\000\128\000\128\000\128\
\\000\128\000\128\000\076\000\076\000\076\000\076\000\076\000\076\
\\000\076\000\077\000\077\000\077\000\077\000\077\000\077\000\077\
\\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\
\\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\
\\000\077\000\077\000\077\000\076\000\076\000\076\000\076\000\077\
\\000\076\000\077\000\077\000\077\000\077\000\077\000\077\000\077\
\\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\
\\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\
\\000\077\000\077\000\077\000\076\000\076\000\076\000\076\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009\
\\000\074\000\074\000\075\000\074\000\074\000\074\000\074\000\074\
\\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\
\\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\
\\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\
\\000\068\000\068\000\070\000\068\000\068\000\068\000\068\000\068\
\\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\
\\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\
\\000\009\000\009\000\009\000\009\000\009\000\009\000\009\000\009"
),
 (3,256,
"\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\001\004\000\247\000\247\001\003\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\001\002\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\252\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\251\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\248\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247\
\\000\247\000\247\000\247\000\247\000\247\000\247\000\247\000\247"
),
 (5,256,
"\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\011\001\005\001\005\001\010\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\008\001\005\001\006\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005\
\\001\005\001\005\001\005\001\005\001\005\001\005\001\005\001\005"
),
 (7,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\012\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (10,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\011\000\011\000\011\000\011\000\011\000\011\000\011\000\011\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (11,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (12,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (13,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\041\000\013\000\039\000\013\000\013\000\024\
\\000\025\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\024\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\022\000\022\000\023\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\017\000\017\000\019\000\017\000\017\000\017\000\017\000\017\
\\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (14,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (15,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (17,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (19,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\021\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\020\000\020\000\020\000\020\000\020\000\020\000\020\000\020\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (21,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\000\000\000\000\013\000\013\000\013\000\013\000\013\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (23,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\000\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\000\000\013\000\013\000\013\000\013\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (24,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\041\000\013\000\042\000\013\000\013\000\024\
\\000\025\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\024\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\022\000\022\000\023\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\017\000\017\000\019\000\017\000\017\000\017\000\017\000\017\
\\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (25,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\041\000\013\000\039\000\013\000\013\000\024\
\\000\037\000\013\000\036\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\024\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\034\000\034\000\035\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\029\000\029\000\031\000\029\000\029\000\029\000\029\000\029\
\\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\
\\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (26,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\027\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (27,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (29,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (31,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\033\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (37,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\000\000\013\000\000\000\013\000\013\000\013\
\\000\038\000\013\000\036\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\034\000\034\000\035\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\029\000\029\000\031\000\029\000\029\000\029\000\029\000\029\
\\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\
\\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (38,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\000\000\013\000\000\000\013\000\013\000\013\
\\000\038\000\013\000\000\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\034\000\034\000\035\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\
\\000\029\000\029\000\031\000\029\000\029\000\029\000\029\000\029\
\\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\
\\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (39,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\040\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (42,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\046\000\000\000\046\000\067\000\046\000\046\000\064\
\\000\065\000\065\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\066\000\065\000\065\000\065\000\065\000\065\000\065\000\065\
\\000\065\000\065\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\046\000\046\000\046\000\046\000\064\
\\000\046\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\046\000\046\000\046\000\046\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\062\000\062\000\063\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\057\000\057\000\059\000\057\000\057\000\057\000\057\000\057\
\\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\
\\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (43,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\044\000\044\000\044\000\044\000\044\000\044\000\044\000\044\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (44,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\045\000\045\000\045\000\045\000\045\000\045\000\045\000\045\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (45,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (46,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\046\000\000\000\046\000\000\000\046\000\046\000\000\
\\000\000\000\000\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\046\000\046\000\046\000\046\000\000\
\\000\046\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\046\000\046\000\046\000\046\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\055\000\055\000\056\000\055\000\055\000\055\000\055\000\055\
\\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\
\\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\
\\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\
\\000\050\000\050\000\052\000\050\000\050\000\050\000\050\000\050\
\\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\
\\000\047\000\047\000\047\000\047\000\047\000\047\000\047\000\047\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (47,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\048\000\048\000\048\000\048\000\048\000\048\000\048\000\048\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (48,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (50,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (52,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\054\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (54,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\000\000\000\000\046\000\046\000\046\000\046\000\046\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (56,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\000\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\046\000\046\000\000\000\046\000\046\000\046\000\046\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (57,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (59,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\061\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (64,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\064\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\000\000\000\000\000\000\000\000\064\
\\000\000\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (68,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (70,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\072\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (72,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\073\000\000\000\013\000\013\000\013\000\013\000\013\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (75,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\073\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\000\000\013\000\013\000\013\000\013\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (77,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\041\000\013\000\104\000\013\000\013\000\078\
\\000\025\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\
\\000\103\000\103\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\013\000\013\000\013\000\013\000\078\
\\000\013\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\022\000\022\000\023\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\017\000\017\000\019\000\017\000\017\000\017\000\017\000\017\
\\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (78,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\041\000\013\000\079\000\013\000\013\000\078\
\\000\025\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\013\000\013\000\013\000\013\000\078\
\\000\013\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\
\\000\078\000\078\000\078\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\022\000\022\000\023\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\017\000\017\000\019\000\017\000\017\000\017\000\017\000\017\
\\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (79,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\083\000\000\000\083\000\067\000\083\000\083\000\101\
\\000\065\000\065\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\102\000\065\000\065\000\065\000\065\000\065\000\065\000\065\
\\000\065\000\065\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\083\000\083\000\083\000\083\000\101\
\\000\083\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\083\000\083\000\083\000\083\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\099\000\099\000\100\000\099\000\099\000\099\000\099\000\099\
\\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\
\\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\
\\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\
\\000\094\000\094\000\096\000\094\000\094\000\094\000\094\000\094\
\\000\094\000\094\000\094\000\094\000\094\000\094\000\094\000\094\
\\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (80,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (81,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (82,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (83,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\083\000\000\000\083\000\000\000\083\000\083\000\000\
\\000\000\000\000\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\083\000\083\000\083\000\083\000\000\
\\000\083\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\083\000\083\000\083\000\083\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\092\000\092\000\093\000\092\000\092\000\092\000\092\000\092\
\\000\092\000\092\000\092\000\092\000\092\000\092\000\092\000\092\
\\000\092\000\092\000\092\000\092\000\092\000\092\000\092\000\092\
\\000\092\000\092\000\092\000\092\000\092\000\092\000\092\000\092\
\\000\087\000\087\000\089\000\087\000\087\000\087\000\087\000\087\
\\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\
\\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (84,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (85,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (87,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (89,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\091\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (91,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\000\000\000\000\083\000\083\000\083\000\083\000\083\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (93,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\000\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\
\\000\083\000\083\000\083\000\000\000\083\000\083\000\083\000\083\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (94,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (96,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\098\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (101,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\101\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\000\000\000\000\000\000\000\000\101\
\\000\000\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\
\\000\101\000\101\000\101\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (104,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\108\000\000\000\108\000\067\000\108\000\108\000\126\
\\000\065\000\065\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\127\000\065\000\065\000\065\000\065\000\065\000\065\000\065\
\\000\065\000\065\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\108\000\108\000\108\000\108\000\126\
\\000\108\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\108\000\108\000\108\000\108\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\124\000\124\000\125\000\124\000\124\000\124\000\124\000\124\
\\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\
\\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\
\\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\
\\000\119\000\119\000\121\000\119\000\119\000\119\000\119\000\119\
\\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\119\
\\000\105\000\105\000\105\000\105\000\105\000\105\000\105\000\105\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (105,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (106,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (107,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (108,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\108\000\000\000\108\000\000\000\108\000\108\000\000\
\\000\000\000\000\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\108\000\108\000\108\000\108\000\000\
\\000\108\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\108\000\108\000\108\000\108\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\117\000\117\000\118\000\117\000\117\000\117\000\117\000\117\
\\000\117\000\117\000\117\000\117\000\117\000\117\000\117\000\117\
\\000\117\000\117\000\117\000\117\000\117\000\117\000\117\000\117\
\\000\117\000\117\000\117\000\117\000\117\000\117\000\117\000\117\
\\000\112\000\112\000\114\000\112\000\112\000\112\000\112\000\112\
\\000\112\000\112\000\112\000\112\000\112\000\112\000\112\000\112\
\\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (109,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (110,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\111\000\111\000\111\000\111\000\111\000\111\000\111\000\111\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (112,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (114,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\116\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (116,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\000\000\000\000\108\000\108\000\108\000\108\000\108\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (118,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\000\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\
\\000\108\000\108\000\108\000\000\000\108\000\108\000\108\000\108\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (119,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (121,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\123\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (126,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\126\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\000\000\000\000\000\000\000\000\126\
\\000\000\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\
\\000\126\000\126\000\126\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (128,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\129\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\131\000\000\
\\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\
\\000\130\000\130\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\000\000\000\000\000\000\000\000\130\
\\000\000\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (131,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\
\\000\132\000\132\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\131\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (132,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\
\\000\132\000\132\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\132\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (133,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\129\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\131\000\000\
\\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\
\\000\130\000\130\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\000\000\000\000\000\000\000\000\130\
\\000\000\000\129\000\137\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\134\000\129\000\129\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (134,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\
\\000\135\000\135\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\135\000\135\000\135\000\135\000\135\000\135\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\136\
\\000\000\000\135\000\135\000\135\000\135\000\135\000\135\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (135,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\129\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\
\\000\135\000\135\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\135\000\135\000\135\000\135\000\135\000\135\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\000\000\000\000\000\000\000\000\135\
\\000\000\000\135\000\135\000\135\000\135\000\135\000\135\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (137,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\
\\000\139\000\139\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\138\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (139,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\129\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\
\\000\139\000\139\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\000\000\000\000\000\000\000\000\139\
\\000\000\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\
\\000\129\000\129\000\129\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (140,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\041\000\013\000\039\000\013\000\013\000\024\
\\000\165\000\013\000\151\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\024\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\149\000\149\000\150\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\144\000\144\000\146\000\144\000\144\000\144\000\144\000\144\
\\000\144\000\144\000\144\000\144\000\144\000\144\000\144\000\144\
\\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (141,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (142,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\143\000\143\000\143\000\143\000\143\000\143\000\143\000\143\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (144,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (146,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\148\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\147\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\147\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\147\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\147\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\147\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\147\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\147\000\147\000\147\000\147\000\147\000\147\000\147\000\147\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (151,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\152\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (152,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\153\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (153,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\154\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (154,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\155\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (155,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\156\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\156\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (156,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\163\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\163\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\
\\000\157\000\157\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (157,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\158\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\158\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\
\\000\157\000\157\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (158,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\158\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\158\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\160\000\000\000\000\000\000\000\000\000\000\
\\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\
\\000\159\000\159\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (159,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\162\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\162\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\160\000\000\000\000\000\000\000\000\000\000\
\\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\
\\000\159\000\159\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (160,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\161\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (162,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\162\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\162\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\160\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (163,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\163\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\163\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\160\000\000\000\000\000\000\000\000\000\000\
\\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\
\\000\164\000\164\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (164,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\158\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\158\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\160\000\000\000\000\000\000\000\000\000\000\
\\000\164\000\164\000\164\000\164\000\164\000\164\000\164\000\164\
\\000\164\000\164\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (165,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\000\000\013\000\000\000\013\000\013\000\013\
\\000\166\000\013\000\036\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\149\000\149\000\150\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\144\000\144\000\146\000\144\000\144\000\144\000\144\000\144\
\\000\144\000\144\000\144\000\144\000\144\000\144\000\144\000\144\
\\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (166,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\000\000\013\000\000\000\013\000\013\000\013\
\\000\166\000\013\000\000\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\149\000\149\000\150\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\
\\000\144\000\144\000\146\000\144\000\144\000\144\000\144\000\144\
\\000\144\000\144\000\144\000\144\000\144\000\144\000\144\000\144\
\\000\141\000\141\000\141\000\141\000\141\000\141\000\141\000\141\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (167,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\000\000\171\000\235\000\171\000\171\000\171\
\\000\222\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\204\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\202\000\202\000\203\000\202\000\202\000\202\000\202\000\202\
\\000\202\000\202\000\202\000\202\000\202\000\202\000\202\000\202\
\\000\202\000\202\000\202\000\202\000\202\000\202\000\202\000\202\
\\000\202\000\202\000\202\000\202\000\202\000\202\000\202\000\202\
\\000\197\000\197\000\199\000\197\000\197\000\197\000\197\000\197\
\\000\197\000\197\000\197\000\197\000\197\000\197\000\197\000\197\
\\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (168,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (169,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (170,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (171,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\041\000\171\000\039\000\171\000\171\000\182\
\\000\183\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\182\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\180\000\180\000\181\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\175\000\175\000\177\000\175\000\175\000\175\000\175\000\175\
\\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\
\\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (172,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (173,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\174\000\174\000\174\000\174\000\174\000\174\000\174\000\174\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (175,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (177,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\179\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (179,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\000\000\000\000\171\000\171\000\171\000\171\000\171\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (181,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\000\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\000\000\171\000\171\000\171\000\171\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (182,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\041\000\171\000\196\000\171\000\171\000\182\
\\000\183\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\182\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\180\000\180\000\181\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\175\000\175\000\177\000\175\000\175\000\175\000\175\000\175\
\\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\
\\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (183,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\041\000\171\000\039\000\171\000\171\000\182\
\\000\194\000\171\000\036\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\182\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\192\000\192\000\193\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\187\000\187\000\189\000\187\000\187\000\187\000\187\000\187\
\\000\187\000\187\000\187\000\187\000\187\000\187\000\187\000\187\
\\000\184\000\184\000\184\000\184\000\184\000\184\000\184\000\184\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (184,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (185,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (187,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\188\000\188\000\188\000\188\000\188\000\188\000\188\000\188\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (189,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\191\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\190\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\190\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\190\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\190\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\190\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\190\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\190\000\190\000\190\000\190\000\190\000\190\000\190\000\190\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (194,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\000\000\171\000\000\000\171\000\171\000\171\
\\000\195\000\171\000\036\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\192\000\192\000\193\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\187\000\187\000\189\000\187\000\187\000\187\000\187\000\187\
\\000\187\000\187\000\187\000\187\000\187\000\187\000\187\000\187\
\\000\184\000\184\000\184\000\184\000\184\000\184\000\184\000\184\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (195,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\000\000\171\000\000\000\171\000\171\000\171\
\\000\195\000\171\000\000\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\192\000\192\000\193\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\192\000\192\000\192\000\192\000\192\000\192\000\192\000\192\
\\000\187\000\187\000\189\000\187\000\187\000\187\000\187\000\187\
\\000\187\000\187\000\187\000\187\000\187\000\187\000\187\000\187\
\\000\184\000\184\000\184\000\184\000\184\000\184\000\184\000\184\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (196,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\046\000\000\000\046\000\040\000\046\000\046\000\064\
\\000\000\000\000\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\066\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\046\000\046\000\046\000\046\000\064\
\\000\046\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\046\000\046\000\046\000\046\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\062\000\062\000\063\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\057\000\057\000\059\000\057\000\057\000\057\000\057\000\057\
\\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\
\\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (197,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\198\000\198\000\198\000\198\000\198\000\198\000\198\000\198\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (199,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\201\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\200\000\200\000\200\000\200\000\200\000\200\000\200\000\200\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (204,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\041\000\171\000\039\000\171\000\171\000\182\
\\000\183\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\182\
\\000\171\000\205\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\180\000\180\000\181\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\175\000\175\000\177\000\175\000\175\000\175\000\175\000\175\
\\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\
\\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (205,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\041\000\171\000\196\000\171\000\171\000\182\
\\000\183\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\182\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\206\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\180\000\180\000\181\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\175\000\175\000\177\000\175\000\175\000\175\000\175\000\175\
\\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\
\\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (206,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\041\000\171\000\207\000\171\000\171\000\182\
\\000\183\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\182\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\180\000\180\000\181\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\
\\000\175\000\175\000\177\000\175\000\175\000\175\000\175\000\175\
\\000\175\000\175\000\175\000\175\000\175\000\175\000\175\000\175\
\\000\172\000\172\000\172\000\172\000\172\000\172\000\172\000\172\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (207,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\046\000\000\000\046\000\040\000\046\000\046\000\064\
\\000\208\000\000\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\066\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\046\000\046\000\046\000\046\000\046\000\046\
\\000\046\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\046\000\046\000\046\000\046\000\064\
\\000\046\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\
\\000\064\000\064\000\064\000\046\000\046\000\046\000\046\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\062\000\062\000\063\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\
\\000\057\000\057\000\059\000\057\000\057\000\057\000\057\000\057\
\\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\
\\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (208,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\208\000\221\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\208\000\208\000\208\000\208\000\000\000\208\000\208\000\208\
\\000\208\000\220\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\219\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\217\000\217\000\218\000\217\000\217\000\217\000\217\000\217\
\\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\
\\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\
\\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\
\\000\212\000\212\000\214\000\212\000\212\000\212\000\212\000\212\
\\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\
\\000\209\000\209\000\209\000\209\000\209\000\209\000\209\000\209\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (209,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\210\000\210\000\210\000\210\000\210\000\210\000\210\000\210\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (210,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (211,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (212,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\213\000\213\000\213\000\213\000\213\000\213\000\213\000\213\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (214,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\216\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (216,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\000\000\000\000\208\000\208\000\208\000\208\000\208\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (218,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\000\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\208\000\000\000\208\000\208\000\208\000\208\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (219,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\208\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\
\\000\208\000\208\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\208\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\208\000\000\
\\000\000\000\000\000\000\000\000\000\208\000\000\000\000\000\000\
\\000\000\000\000\000\208\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (222,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\041\000\171\000\039\000\171\000\171\000\182\
\\000\233\000\171\000\000\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\182\
\\000\171\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\
\\000\182\000\182\000\182\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\231\000\231\000\232\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\226\000\226\000\228\000\226\000\226\000\226\000\226\000\226\
\\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\
\\000\223\000\223\000\223\000\223\000\223\000\223\000\223\000\223\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (223,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\224\000\224\000\224\000\224\000\224\000\224\000\224\000\224\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (224,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (226,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\227\000\227\000\227\000\227\000\227\000\227\000\227\000\227\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (228,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\230\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\229\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\229\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\229\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\229\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\229\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\229\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\229\000\229\000\229\000\229\000\229\000\229\000\229\000\229\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (233,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\000\000\171\000\000\000\171\000\171\000\171\
\\000\234\000\171\000\036\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\231\000\231\000\232\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\226\000\226\000\228\000\226\000\226\000\226\000\226\000\226\
\\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\
\\000\223\000\223\000\223\000\223\000\223\000\223\000\223\000\223\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (234,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\171\000\000\000\171\000\000\000\171\000\171\000\171\
\\000\234\000\171\000\000\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\
\\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\231\000\231\000\232\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\231\000\231\000\231\000\231\000\231\000\231\000\231\000\231\
\\000\226\000\226\000\228\000\226\000\226\000\226\000\226\000\226\
\\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\
\\000\223\000\223\000\223\000\223\000\223\000\223\000\223\000\223\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (235,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\240\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\236\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (236,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\237\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (237,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\238\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (238,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\239\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (239,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\208\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (240,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\240\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (241,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\013\000\242\000\013\000\039\000\013\000\013\000\024\
\\000\025\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\013\000\013\000\013\000\013\000\013\000\013\000\013\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\024\
\\000\013\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\
\\000\024\000\024\000\024\000\013\000\013\000\013\000\013\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\022\000\022\000\023\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\
\\000\017\000\017\000\019\000\017\000\017\000\017\000\017\000\017\
\\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\
\\000\014\000\014\000\014\000\014\000\014\000\014\000\014\000\014\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (245,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\246\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (248,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\249\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (249,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\250\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (251,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\250\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (252,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\255\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\001\001\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\253\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (253,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\254\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (254,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\255\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (256,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\255\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (259,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\004\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (262,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\001\007\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (264,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\009\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
 (266,256,
"\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\001\011\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000"
),
(0, 0, "")]
fun f(n, i, x) = (n, Vector.tabulate(i, decode x))
val s = map f (rev (tl (rev s)))
exception LexHackingError
fun look ((j,x)::r, i) = if i = j then x else look(r, i)
  | look ([], i) = raise LexHackingError
fun g {fin=x, trans=i} = {fin=x, trans=look(s,i)}
in Vector.fromList(map g
[{fin = [], trans = 0},
{fin = [], trans = 1},
{fin = [], trans = 1},
{fin = [], trans = 3},
{fin = [], trans = 3},
{fin = [], trans = 5},
{fin = [], trans = 5},
{fin = [(N 897)], trans = 7},
{fin = [(N 897)], trans = 7},
{fin = [(N 857)], trans = 0},
{fin = [(N 857)], trans = 10},
{fin = [], trans = 11},
{fin = [], trans = 12},
{fin = [(N 852)], trans = 13},
{fin = [], trans = 14},
{fin = [], trans = 15},
{fin = [], trans = 12},
{fin = [], trans = 17},
{fin = [], trans = 12},
{fin = [], trans = 19},
{fin = [], trans = 12},
{fin = [], trans = 21},
{fin = [], trans = 12},
{fin = [], trans = 23},
{fin = [(N 852)], trans = 24},
{fin = [(N 852)], trans = 25},
{fin = [], trans = 26},
{fin = [], trans = 27},
{fin = [], trans = 12},
{fin = [], trans = 29},
{fin = [], trans = 12},
{fin = [], trans = 31},
{fin = [], trans = 12},
{fin = [], trans = 21},
{fin = [], trans = 12},
{fin = [], trans = 23},
{fin = [(N 755)], trans = 0},
{fin = [], trans = 37},
{fin = [], trans = 38},
{fin = [], trans = 39},
{fin = [(N 506)], trans = 39},
{fin = [(N 656)], trans = 0},
{fin = [], trans = 42},
{fin = [], trans = 43},
{fin = [], trans = 44},
{fin = [], trans = 45},
{fin = [(N 254)], trans = 46},
{fin = [], trans = 47},
{fin = [], trans = 48},
{fin = [], trans = 45},
{fin = [], trans = 50},
{fin = [], trans = 45},
{fin = [], trans = 52},
{fin = [], trans = 45},
{fin = [], trans = 54},
{fin = [], trans = 45},
{fin = [], trans = 56},
{fin = [], trans = 57},
{fin = [], trans = 45},
{fin = [], trans = 59},
{fin = [], trans = 45},
{fin = [], trans = 54},
{fin = [], trans = 45},
{fin = [], trans = 56},
{fin = [(N 254)], trans = 64},
{fin = [(N 406)], trans = 0},
{fin = [(N 254)], trans = 0},
{fin = [(N 406),(N 506)], trans = 39},
{fin = [(N 857)], trans = 68},
{fin = [], trans = 12},
{fin = [(N 857)], trans = 70},
{fin = [], trans = 12},
{fin = [], trans = 72},
{fin = [(N 548)], trans = 0},
{fin = [(N 857)], trans = 12},
{fin = [(N 857)], trans = 75},
{fin = [(N 852),(N 857)], trans = 13},
{fin = [(N 852),(N 857)], trans = 77},
{fin = [(N 852)], trans = 78},
{fin = [], trans = 79},
{fin = [], trans = 80},
{fin = [], trans = 81},
{fin = [], trans = 82},
{fin = [(N 254),(N 305)], trans = 83},
{fin = [], trans = 84},
{fin = [], trans = 85},
{fin = [], trans = 82},
{fin = [], trans = 87},
{fin = [], trans = 82},
{fin = [], trans = 89},
{fin = [], trans = 82},
{fin = [], trans = 91},
{fin = [], trans = 82},
{fin = [], trans = 93},
{fin = [], trans = 94},
{fin = [], trans = 82},
{fin = [], trans = 96},
{fin = [], trans = 82},
{fin = [], trans = 91},
{fin = [], trans = 82},
{fin = [], trans = 93},
{fin = [(N 254),(N 305)], trans = 101},
{fin = [(N 254),(N 305)], trans = 0},
{fin = [(N 852)], trans = 77},
{fin = [], trans = 104},
{fin = [], trans = 105},
{fin = [], trans = 106},
{fin = [], trans = 107},
{fin = [(N 305)], trans = 108},
{fin = [], trans = 109},
{fin = [], trans = 110},
{fin = [], trans = 107},
{fin = [], trans = 112},
{fin = [], trans = 107},
{fin = [], trans = 114},
{fin = [], trans = 107},
{fin = [], trans = 116},
{fin = [], trans = 107},
{fin = [], trans = 118},
{fin = [], trans = 119},
{fin = [], trans = 107},
{fin = [], trans = 121},
{fin = [], trans = 107},
{fin = [], trans = 116},
{fin = [], trans = 107},
{fin = [], trans = 118},
{fin = [(N 305)], trans = 126},
{fin = [(N 305)], trans = 0},
{fin = [(N 541),(N 857)], trans = 128},
{fin = [(N 541)], trans = 0},
{fin = [(N 541)], trans = 128},
{fin = [], trans = 131},
{fin = [(N 519)], trans = 132},
{fin = [(N 541),(N 857)], trans = 133},
{fin = [(N 541)], trans = 134},
{fin = [(N 541)], trans = 135},
{fin = [], trans = 134},
{fin = [(N 541)], trans = 137},
{fin = [], trans = 137},
{fin = [(N 541)], trans = 139},
{fin = [(N 852),(N 857)], trans = 140},
{fin = [], trans = 141},
{fin = [], trans = 142},
{fin = [], trans = 12},
{fin = [], trans = 144},
{fin = [], trans = 12},
{fin = [], trans = 146},
{fin = [], trans = 12},
{fin = [], trans = 21},
{fin = [], trans = 12},
{fin = [], trans = 23},
{fin = [(N 509)], trans = 151},
{fin = [], trans = 152},
{fin = [], trans = 153},
{fin = [], trans = 154},
{fin = [], trans = 155},
{fin = [], trans = 156},
{fin = [], trans = 157},
{fin = [], trans = 158},
{fin = [], trans = 159},
{fin = [], trans = 160},
{fin = [(N 107)], trans = 0},
{fin = [], trans = 162},
{fin = [], trans = 163},
{fin = [], trans = 164},
{fin = [], trans = 165},
{fin = [], trans = 166},
{fin = [(N 855),(N 857)], trans = 167},
{fin = [], trans = 168},
{fin = [], trans = 169},
{fin = [], trans = 170},
{fin = [(N 852)], trans = 171},
{fin = [], trans = 172},
{fin = [], trans = 173},
{fin = [], trans = 170},
{fin = [], trans = 175},
{fin = [], trans = 170},
{fin = [], trans = 177},
{fin = [], trans = 170},
{fin = [], trans = 179},
{fin = [], trans = 170},
{fin = [], trans = 181},
{fin = [(N 852)], trans = 182},
{fin = [(N 852)], trans = 183},
{fin = [], trans = 184},
{fin = [], trans = 185},
{fin = [], trans = 170},
{fin = [], trans = 187},
{fin = [], trans = 170},
{fin = [], trans = 189},
{fin = [], trans = 170},
{fin = [], trans = 179},
{fin = [], trans = 170},
{fin = [], trans = 181},
{fin = [], trans = 194},
{fin = [], trans = 195},
{fin = [], trans = 196},
{fin = [], trans = 197},
{fin = [], trans = 170},
{fin = [], trans = 199},
{fin = [], trans = 170},
{fin = [], trans = 179},
{fin = [], trans = 170},
{fin = [], trans = 181},
{fin = [(N 852)], trans = 204},
{fin = [(N 852)], trans = 205},
{fin = [(N 852)], trans = 206},
{fin = [], trans = 207},
{fin = [], trans = 208},
{fin = [], trans = 209},
{fin = [], trans = 210},
{fin = [], trans = 211},
{fin = [], trans = 212},
{fin = [], trans = 211},
{fin = [], trans = 214},
{fin = [], trans = 211},
{fin = [], trans = 216},
{fin = [], trans = 211},
{fin = [], trans = 218},
{fin = [], trans = 219},
{fin = [(N 45)], trans = 0},
{fin = [(N 91)], trans = 0},
{fin = [(N 852)], trans = 222},
{fin = [], trans = 223},
{fin = [], trans = 224},
{fin = [], trans = 170},
{fin = [], trans = 226},
{fin = [], trans = 170},
{fin = [], trans = 228},
{fin = [], trans = 170},
{fin = [], trans = 179},
{fin = [], trans = 170},
{fin = [], trans = 181},
{fin = [], trans = 233},
{fin = [], trans = 234},
{fin = [(N 855)], trans = 235},
{fin = [], trans = 236},
{fin = [], trans = 237},
{fin = [], trans = 238},
{fin = [], trans = 239},
{fin = [(N 855)], trans = 240},
{fin = [(N 852),(N 857)], trans = 241},
{fin = [(N 551),(N 656)], trans = 0},
{fin = [(N 548),(N 857)], trans = 0},
{fin = [(N 553),(N 857)], trans = 0},
{fin = [(N 558),(N 857)], trans = 245},
{fin = [(N 558)], trans = 0},
{fin = [(N 877)], trans = 0},
{fin = [(N 877)], trans = 248},
{fin = [], trans = 249},
{fin = [(N 864)], trans = 0},
{fin = [(N 877)], trans = 251},
{fin = [(N 877)], trans = 252},
{fin = [], trans = 253},
{fin = [], trans = 254},
{fin = [(N 872)], trans = 0},
{fin = [], trans = 256},
{fin = [(N 875)], trans = 0},
{fin = [(N 864),(N 877)], trans = 0},
{fin = [(N 877),(N 882)], trans = 259},
{fin = [(N 882)], trans = 0},
{fin = [(N 895)], trans = 0},
{fin = [(N 895)], trans = 262},
{fin = [(N 885)], trans = 0},
{fin = [(N 895)], trans = 264},
{fin = [(N 888)], trans = 0},
{fin = [(N 893),(N 895)], trans = 266},
{fin = [(N 893)], trans = 0},
{fin = [(N 897)], trans = 7}])
end
structure StartStates =
    struct
    datatype yystartstate = STARTSTATE of int

(* start state definitions *)

val INITIAL = STARTSTATE 1;
val comment = STARTSTATE 5;
val recover_stash = STARTSTATE 7;
val string = STARTSTATE 3;

end
type result = UserDeclarations.lexresult
    exception LexerError (* raised if illegal leaf action tried *)
end

type int = Int.int
fun makeLexer (yyinput: int -> string) =
let    val yygone0:int=0
    val yyb = ref "\n"         (* buffer *)
    val yybl: int ref = ref 1        (*buffer length *)
    val yybufpos: int ref = ref 1        (* location of next character to use *)
    val yygone: int ref = ref yygone0    (* position in file of beginning of buffer *)
    val yydone = ref false        (* eof found yet? *)
    val yybegin: int ref = ref 1        (*Current 'start state' for lexer *)

    val YYBEGIN = fn (Internal.StartStates.STARTSTATE x) =>
         yybegin := x

fun lex (yyarg as ({rcopt,nf,r,i,incommentp,stracc,stash} : UserDeclarations.extrastate)) =
let fun continue() : Internal.result =
  let fun scan (s,AcceptingLeaves : Internal.yyfinstate list list,l,i0: int) =
    let fun action (i: int,nil) = raise LexError
    | action (i,nil::l) = action (i-1,l)
    | action (i,(node::acts)::l) =
        case node of
            Internal.N yyk =>
            (let fun yymktext() = String.substring(!yyb,i0,i-i0)
                 val yypos: int = i0+ !yygone
            open UserDeclarations Internal.StartStates
 in (yybufpos := i; case yyk of

            (* Application actions *)

  107 => let val yytext=yymktext() in dolocpragma yyarg yypos yytext; continue() end
| 254 => let val yytext=yymktext() in let val (part1, part2) = cut_qid yytext
  in
    if size part1 = 0 then (BT_Ident yytext, getLoc yyarg yypos yytext)
    else (
      YYBEGIN recover_stash;
      stash := part2;
      (BT_Ident part1, getLoc yyarg yypos yytext)
    )
  end
 end
| 305 => let val yytext=yymktext() in (BT_Ident yytext, getLoc yyarg yypos yytext)
 end
| 406 => let val yytext=yymktext() in raise LEX_ERR ("Bad qualified ident", getLoc yyarg yypos yytext)
 end
| 45 => let val yytext=yymktext() in (BT_Ident yytext, getLoc yyarg yypos yytext)
 end
| 506 => let val yytext=yymktext() in raise LEX_ERR ("Bad qualified ident", getLoc yyarg yypos yytext)
 end
| 509 => (YYBEGIN comment; incommentp := 1; continue())
| 519 => let val yytext=yymktext() in let val loc = getLoc yyarg yypos yytext
   in
     (BT_DecimalFraction (parse_fraction (yytext, loc)), loc)
   end end
| 541 => let val yytext=yymktext() in let val loc = getLoc yyarg yypos yytext
     in
       (BT_Numeral(parse_numeric_literal(yytext,loc)), loc)
     end end
| 548 => let val yytext=yymktext() in let
                  val s = LocP(nf,!r,yypos - !(#i yyarg))
                in
                  stracc := ("", [], s,
                             (yytext, matching_strdelim yytext));
                  YYBEGIN string;
                  continue()
                end end
| 551 => (let val s = LocP(nf,!r,yypos - !(#i yyarg))
                 in
                     stracc := ("", [], s, ("#\"", "\""));
                     YYBEGIN string;
                     check_char (continue())
                 end)
| 553 => (continue())
| 558 => let val yytext=yymktext() in let val i = #i yyarg
                     in
                       r := !r + 1;
                       i := yypos + size yytext;
                       continue()
                     end end
| 656 => let val yytext=yymktext() in (* needed to cope with stuff smashed up against char literals *)
  let
    val s = LocP(nf,!r,yypos - !(#i yyarg))
  in
    stracc := (String.substring(yytext,0,size yytext - 1), [], s, ("\"", "\""));
    YYBEGIN string;
    continue()
  end
 end
| 755 => let val yytext=yymktext() in YYBEGIN comment;
                                 incommentp := 1;
                                 (BT_Ident(String.substring(yytext,0,size yytext - 2)),
                                  getLoc yyarg yypos yytext) end
| 852 => let val yytext=yymktext() in (BT_Ident yytext, getLoc yyarg yypos yytext) end
| 855 => let val yytext=yymktext() in (BT_Ident yytext, getLoc yyarg yypos yytext) end
| 857 => let val yytext=yymktext() in raise LEX_ERR("Character \""^yytext^"\" is a lexical error",
                             getLoc yyarg yypos yytext) end
| 864 => let val yytext=yymktext() in let val e = LocP(nf,!r,yypos - !(#i yyarg))
      val (pfx, clist, start, (ldelim,rdelim)) = !stracc
      val str0 = String.concat (List.rev clist)
      val str = MLstring.scanMLstring str0
                 handle MLstring.stringerror(i,s) =>
                        raise LEX_ERR (s ^ " at string character "^
                                       Int.toString i,
                                       Loc(start,e))
  in
    if rdelim = yytext then
      (stracc := ("", [], LocP(0,0,0), ("",""));
       YYBEGIN INITIAL;
       (BT_StrLit{ldelim = pfx^ldelim, contents = str}, Loc(start,e)))
    else
      raise LEX_ERR ("Character \""^yytext^"\" can't occur inside this sort of string literal unescaped", Loc(start,e))
  end  end
| 872 => (let val (pfx, clist, start, delim) = !stracc
                   in
                     stracc := (pfx, "\\\"" :: clist, start, delim);
                     continue()
                   end)
| 875 => (let val (pfx, clist, start, delim) = !stracc
                   in
                     stracc := (pfx, "\\\\" :: clist, start, delim);
                     continue()
                   end)
| 877 => let val yytext=yymktext() in let val (pfx, clist, start, delim) = !stracc
               in
                 stracc := (pfx, yytext :: clist, start, delim);
                 continue()
               end end
| 882 => let val yytext=yymktext() in let val i = #i yyarg
                          val (pfx, clist,start,delim) = !stracc
                      in
                        r := !r + 1;
                        i := yypos + size yytext;
                        stracc := (pfx, "\n" :: clist, start, delim);
                        continue()
                      end end
| 885 => (incommentp := !incommentp - 1;
                  if !incommentp = 0 then YYBEGIN INITIAL else ();
                  continue())
| 888 => (incommentp := !incommentp + 1;
                  continue())
| 893 => let val yytext=yymktext() in let val i = #i yyarg
                     in
                       r := !r + 1;
                       i := yypos + size yytext;
                       continue()
                     end end
| 895 => (continue())
| 897 => let val yytext=yymktext() in let
    val s = LocP(nf,!r,yypos - !(#i yyarg))
  in
    if size (!stash) = 0 then (YYBEGIN INITIAL; continue())
    else
      case String.sub(!stash, 0) of
          #"\"" => (if size (!stash) > 1 then
                      (stash := String.extract (!stash, 2, NONE);
                       (BT_StrLit{ldelim="\"", contents = ""},
                        getLoc yyarg yypos yytext))
                    else (YYBEGIN string; stracc := ("", [], s, ("#\"","\""));
                          continue()))
        | _ =>
          (case size yytext of
               0 => (YYBEGIN INITIAL;
                     (BT_Ident (!stash), getLoc yyarg yypos yytext) before
                     stash := "")
             | 1 => (YYBEGIN string;
                     stracc := ("", [], s, ("\"","\""));
                     (BT_Ident (!stash), getLoc yyarg yypos yytext) before
                     stash := "")
             | n => ((BT_Ident (!stash), getLoc yyarg yypos yytext) before
                     stash := yytext))
  end
 end
| 91 => let val yytext=yymktext() in raise LEX_ERR ("Illegal form for quoted variable", getLoc yyarg yypos yytext)
 end
| _ => raise Internal.LexerError

        ) end )

    val {fin,trans} = Vector.sub (Internal.tab, s)
    val NewAcceptingLeaves = fin::AcceptingLeaves
    in if l = !yybl then
         if trans = #trans(Vector.sub(Internal.tab,0))
           then action(l,NewAcceptingLeaves
) else        let val newchars= if !yydone then "" else yyinput 1024
        in if (String.size newchars)=0
          then (yydone := true;
                if (l=i0) then UserDeclarations.eof yyarg
                          else action(l,NewAcceptingLeaves))
          else (if i0=l then yyb := newchars
             else yyb := String.substring(!yyb,i0,l-i0)^newchars;
             yygone := !yygone+i0;
             yybl := String.size (!yyb);
             scan (s,AcceptingLeaves,l-i0,0))
        end
      else let val NewChar = Char.ord (CharVector.sub (!yyb,l))
        val NewState = Vector.sub (trans, NewChar)
        in if NewState=0 then action(l,NewAcceptingLeaves)
        else scan(NewState,NewAcceptingLeaves,l+1,i0)
    end
    end
(*
    val start= if String.substring(!yyb,!yybufpos-1,1)="\n"
then !yybegin+1 else !yybegin
*)
    in scan(!yybegin (* start *),nil,!yybufpos,!yybufpos)
    end
in continue end
  in lex
  end
end
